#include <xtl/scheduler.hpp>
#include <xtl/fiber.hpp>
#include <xtl/thread.hpp>

#include <xtl/exception.hpp>
#include <xtl/stack.hpp>
#include <unordered_map>
#include <unordered_set>
#include <map>
#include <deque>
#include <cstring>
#include <sys/epoll.h>
#include <ucontext.h>

#if defined(DEBUG_SCHEDULER) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) fprintf(stderr, "scheduler[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "scheduler[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xtl {

struct execution_context {
	ucontext_t ucontext;
	size_t timeout_ms;
	uint64_t timeout_abs;
	int blocker;

	inline execution_context() noexcept : timeout_ms(0) , timeout_abs(0), blocker(-1) {}
};

struct epoll_event : ::epoll_event {
	inline epoll_event(int fd) noexcept {
		memset(this, 0, sizeof(*this));
		events = EPOLLIN | EPOLLOUT | EPOLLERR | EPOLLHUP | EPOLLRDHUP | EPOLLET;
		data.fd = fd;
	}
};

inline int make_epoll() {
	const int fd = ::epoll_create1(EPOLL_CLOEXEC);
	if (fd < 0) {
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	return fd;
}

template<typename T>
union fiber_arg {
public:
	inline constexpr fiber_arg(T * value) noexcept : m_value(value) {}
	inline constexpr fiber_arg(int a, int b) noexcept : m_ints{a, b} {}

	inline T * operator->() const noexcept {
		return m_value;
	}

	inline int operator[](size_t index) const noexcept {
		return m_ints[index];
	}

private:
	T * m_value;
	int m_ints[2];
};

class scheduler_context {
public:
	inline scheduler_context()
	: m_fd(make_epoll())
	, m_current_context(&m_thread_context)
	, m_scheduler_context(create_fiber(nullptr, xtl::function<void()>([this] () { run(); }), 16384))
	, m_active(true)
	{
		fdebug("Created scheduler %p, thread context %p", (void *) this, (void *) &m_thread_context);
	}

	inline ~scheduler_context() noexcept {
		m_active = false;
		fdebug("Destroyed scheduler %p", (void *) this);
	}

	inline void monitor(const xtl::file_descriptor & _fd) {
		const int fd = _fd.value();
		fdebug("monitor(%d)", fd);
		const auto monitored = m_monitored.find(fd);
		if (monitored != m_monitored.end()) {
			return;
		}
		xtl::epoll_event event(fd);
		if (::epoll_ctl(m_fd.value(), EPOLL_CTL_ADD, fd, &event) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		m_monitored.emplace(std::piecewise_construct, std::forward_as_tuple(fd), std::forward_as_tuple());
	}

	inline void release(const xtl::file_descriptor & _fd) {
		const int fd = _fd.value();
		fdebug("release(%d)", fd);
		const auto monitored = m_monitored.find(fd);
		if (monitored == m_monitored.end()) {
			return;
		}
		if (::epoll_ctl(m_fd.value(), EPOLL_CTL_DEL, fd, nullptr) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		for (xtl::execution_context * context : monitored->second) {
			//~ clear_timeout(context);
			m_active_contexts.push_back(context);
		}
		m_monitored.erase(monitored);
	}

	inline void clear_timeout(xtl::execution_context * context) {
		//~ context->timeout_ms = 0;
		//~ const uint64_t timeout = std::exchange(context->timeout_abs, 0);
		//~ if (timeout > 0) {
			//~ const auto contexts = m_timeouts.find(timeout);
			//~ if (contexts != m_timeouts.end()) {
				//~ contexts->second.erase(context);
				//~ if (contexts->second.empty()) {
					//~ m_timeouts.erase(contexts);
				//~ }
			//~ }
		//~ }
	}

	inline void set_timeout(const xtl::file_descriptor & _fd, size_t timeout_ms) {
		//~ const int fd = _fd.value();
		//~ const auto monitored = m_monitored.find(fd);
		//~ if (monitored == m_monitored.end()) {
			//~ // FDs are exclusively owned by a single resource
			//~ throw xtl::logic_exception(__FILE__, __LINE__, "Invalid FD");
		//~ }
		//~ if (timeout_ms > 0) {
			//~ const uint64_t timeout_abs = now() + timeout_ms;
			//~ auto & timeouts = m_timeouts[timeout_abs];
			//~ for (xtl::execution_context * context : monitored->second) {
				//~ clear_timeout(context);
				//~ context->timeout_ms = timeout_ms;
				//~ context->timeout_abs = timeout_abs;
				//~ timeouts.emplace(context);
			//~ }
		//~ } else {
			//~ for (xtl::execution_context * context : monitored->second) {
				//~ clear_timeout(context);
			//~ }
		//~ }
	}

	inline void wait_for(const xtl::file_descriptor & _fd) {
		const int fd = _fd.value();
		fdebug("wait_for(%d) (cc %p)", fd, (void *) m_current_context);
		m_current_context->blocker = fd;
		const auto monitored = m_monitored.find(fd);
		if (monitored != m_monitored.end()) {
			monitored->second.emplace(m_current_context);
		} else {
			monitor(fd);
			m_monitored[fd].emplace(m_current_context);
		}
		yield();
	}

	inline void yield() {
		if (m_active_contexts.empty()) {
			switch_to(m_scheduler_context);
		}
		execute_active_contexts();
	}

	xtl::execution_context * create_fiber(
		void * fiber,
		xtl::function<void()> && function,
		size_t stack_size
	) {
		fiberdata & fd = m_fibers.emplace(
			std::piecewise_construct,
			std::forward_as_tuple(fiber),
			std::forward_as_tuple(std::move(function), stack_size)
		).first->second;
		if (::getcontext(&fd.context.ucontext) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		fd.context.ucontext.uc_stack.ss_sp = fd.stack.get();
		fd.context.ucontext.uc_stack.ss_size = fd.stack.size();
		fd.context.ucontext.uc_link = nullptr;
		const xtl::fiber_arg<fiberdata> arg(&fd);
		::makecontext(&fd.context.ucontext, reinterpret_cast<void (*)()>(trampoline), 2, arg[0], arg[1]);
		m_active_contexts.push_back(&fd.context);
		return &fd.context;
	}

	inline void destroy_fiber(void * fiber) {
		fdebug("destroy_fiber(%p)", fiber);
		for (auto it = m_fibers.find(fiber); it != m_fibers.end();) {
			xtl::execution_context * context = &it->second.context;
			const int blocker = context->blocker;
			if (blocker >= 0) {
				m_monitored[blocker].erase(context);
			}
			deactivate_context(context);
			m_dead_fibers.push_back(fiber);
			return;
		}
	}

private:
	struct fiberdata {
		inline fiberdata(
			xtl::function<void()> && function,
			size_t stack_size
		)
		: function(std::move(function))
		, stack(stack_size)
		{
			fdebug("Created fiberdata %p", (void *) &context);
		}

		inline ~fiberdata() noexcept {
			fdebug("Destroying fiberdata %p", (void *) &context);
		}

		xtl::function<void()> function;
		xtl::execution_context context;
		xtl::stack stack;
	};

	xtl::file_descriptor m_fd;
	xtl::execution_context * m_current_context;
	std::unordered_map<void *, fiberdata> m_fibers;
	std::unordered_map<int, std::unordered_set<xtl::execution_context *>> m_monitored;
	//~ std::map<uint64_t, std::unordered_set<xtl::execution_context *>> m_timeouts;
	std::deque<xtl::execution_context *> m_active_contexts;
	std::deque<void *> m_dead_fibers;
	xtl::execution_context * m_scheduler_context;
	xtl::execution_context m_thread_context;
	bool m_active;

	static void trampoline(int a, int b) {
		try {
			xtl::fiber_arg<fiberdata>(a, b)->function();
		} catch (const std::exception & e) {
			fprintf(stderr, "%s\n", e.what());
		}
		xtl::yield();
		::abort(); // should never happen
	}

	void deactivate_context(xtl::execution_context * context) {
		for (auto it = m_active_contexts.begin(); it != m_active_contexts.end();) {
			if (*it == context) {
				it = m_active_contexts.erase(it);
			} else {
				++it;
			}
		}
	}

	void run() {
		while (m_active) {
			execute_active_contexts();
			reap_dead_fibers();
			wait_for_events();
		}
		switch_to(&m_thread_context);
	}

	void reap_dead_fibers() {
		while (!m_dead_fibers.empty()) {
			void * fiber = m_dead_fibers.front();
			fdebug("Reaping fiber %p", fiber);
			m_dead_fibers.pop_front();
			m_fibers.erase(fiber);
		}
	}

	void execute_active_contexts() {
		while (!m_active_contexts.empty()) {
			xtl::execution_context * context = m_active_contexts.front();
			m_active_contexts.pop_front();
			switch_to(context);
		}
	}

	inline void switch_to(xtl::execution_context * context) {
		fdebug("Switching from %p to %p", (void *) m_current_context, (void *) context);
		if (::swapcontext(&std::exchange(m_current_context, context)->ucontext, &context->ucontext) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
	}

	inline size_t now() {
		struct timespec res;
		if (clock_gettime(CLOCK_MONOTONIC, &res) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		return (uint64_t(res.tv_sec) * 1000) + (res.tv_nsec / 1000000);
	}

	void wait_for_events() {
		const size_t max_events = 16;
		struct ::epoll_event events[max_events];
		for (;;) {
			int timeout = -1;
			//~ if (!m_timeouts.empty()) {
				//~ const auto _now = now();
				//~ const auto next = m_timeouts.begin()->first;
				//~ timeout = (next > _now) ? _now - next : 0;
			//~ }
			const int result = ::epoll_wait(m_fd.value(), events, max_events, timeout);
			if (result > 0) {
				for (int i = 0; i < result; ++i) {
					const auto fd = events[i].data.fd;
					const auto monitored = m_monitored.find(fd);
					if (monitored != m_monitored.end()) {
						for (xtl::execution_context * context : monitored->second) {
							fdebug("unblocking %p", (void *) context);
							// TODO: restart timeout on context switch
							m_active_contexts.push_back(context);
						}
					}
				}
			} else if (result == 0) {
				//~ const uint64_t _now = now();
				//~ for (auto it = m_timeouts.begin(); it != m_timeouts.end() && _now > it->first;) {
					//~ for (xtl::execution_context * context : it->second) {
						//~ // TODO: restart timeout on context switch
						//~ m_active_contexts.push_back(context);
					//~ }
					//~ it = m_timeouts.erase(it);
				//~ }
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::system_exception(__FILE__, __LINE__);
			}
			return;
		}
	}
};

pthread_key_t g_sched_key;
pthread_once_t g_sched_once = PTHREAD_ONCE_INIT;

scheduler_context & current_scheduler() {
	pthread_once(&g_sched_once, [] () {
		pthread_key_create(&g_sched_key, [] (void * ptr) {
			delete static_cast<scheduler_context *>(ptr);
		});
	});
	scheduler_context * scheduler = static_cast<scheduler_context *>(pthread_getspecific(g_sched_key));
	if (scheduler == nullptr) {
		scheduler = new scheduler_context();
		pthread_setspecific(g_sched_key, scheduler);
	}
	return *scheduler;
}

void scheduler::monitor(const xtl::file_descriptor & fd) {
	current_scheduler().monitor(fd);
}

void scheduler::release(const xtl::file_descriptor & fd) {
	current_scheduler().release(fd);
}

void scheduler::set_timeout(const xtl::file_descriptor & fd, size_t timeout_ms) {
	current_scheduler().set_timeout(fd, timeout_ms);
}

void scheduler::wait_for(const xtl::file_descriptor & fd) {
	current_scheduler().wait_for(fd);
}

void yield() {
	current_scheduler().yield();
}

void scheduler::create_fiber(void * fiber, xtl::function<void()> && function, size_t stack_size) {
	current_scheduler().create_fiber(fiber, std::move(function), stack_size);
}

void scheduler::destroy_fiber(void * fiber) {
	current_scheduler().destroy_fiber(fiber);
}

} // namespace xtl
